#include "binary_belief_propagation.h"

#include "definitions.h"
#include "binary_factor_graph.h"
#include "factor.h"
#include "binary_variable.h"
#include "message.h"
#include "edge.h"


BinaryBeliefPropagation::BinaryBeliefPropagation(BinaryFactorGraph *fg) {
  fg_ = fg;
  iteration_ = 0;
  verbose_ = false;
  messages_sent_ = 0;
  converged_ = false;
  SetSchedule(SYNCHRONOUS);
}


BinaryBeliefPropagation::~BinaryBeliefPropagation() {
  /*
  for (uint i = 0; i < messages_->size(); i++) {
    delete messages_->at(i);
  }
  */
}


void BinaryBeliefPropagation::AssignVariablesByBelief() {
  int num_vars = static_cast<int>(fg_->Variables()->size());

  for (int i = 0; i < num_vars; i++) {
    BinaryVariable *var = fg_->Variables()->at(i);
    double belief = var->Belief();
    if (belief >= 0) {
      var->AssignValue(1);
    } else if (belief < 0) {
      var->AssignValue(0);
    }
  }
}


// Go through each set of variables where exactly one in the set
// is constrained to be on, and choose the one with the largest
// belief.  If there are overlapping scopes, this DOES NOT guarantee
// that we'll have a legal assignment.  It also assumes each variable
// belongs to at least one hard-constrained factor.  TODO: relax this.
void BinaryBeliefPropagation::AssignVariablesByHardConstraints() {
  for (int i = 0; i < static_cast<int>(fg_->Factors()->size()); i++) {
    Factor *factor = fg_->Factors()->at(i);
    if (factor->FactorClass() == HARD_1_OF_N) {
      vector<BinaryVariable *> *scope = factor->Scope();
      
      int max_idx = -1;
      double max_belief = -HUGE_VAL;
      for (int j = 0; j < static_cast<int>(scope->size()); j++) {
	double belief_j = scope->at(j)->Belief();
	if (belief_j > max_belief) {
	  max_idx = j;
	  max_belief = belief_j;
	}

	// Assign everything to value 0 in the first pass, then
	// we'll come back and replace the assignment for the max_idx
	scope->at(j)->AssignValue(0);
      }

      // Assign the max to take on value 1
      scope->at(max_idx)->AssignValue(1);
    }
  }
}

void BinaryBeliefPropagation::InitMessagePassing() {
  if (schedule_ == SYNCHRONOUS) {
    cerr << "Initialized messages for " << fg_->Variables()->size() 
	 << " variable and " << fg_->Factors()->size() << " factors" << endl;
  }
}


double BinaryBeliefPropagation::PassMessages() {
  if (verbose_) {
    cerr << "Iteration " << iteration_ << endl;
  } else {
    //cerr << "Iteration " << iteration_ << endl;
    //cerr << "\r" << "Iteration " << iteration_;
  }
  double residual = 0;
  if (schedule_ == SYNCHRONOUS) {
    for (uint i = 0; i < fg_->Factors()->size(); i++) {
      Factor *factor = fg_->Factors()->at(i);
      factor->Touch();
      factor->ComputeAllOutgoingMessages();
      residual += factor->UpdateAllOutgoingMessages();
    }

    for (uint i = 0; i < fg_->Variables()->size(); i++) {
      BinaryVariable *variable = fg_->Variables()->at(i);
      variable->Touch();
      variable->ComputeAllOutgoingMessages();
      residual += variable->UpdateAllOutgoingMessages();
    }
  } else {
    cerr << "ERROR: Message schedule not implemented (" << schedule_ << ")" << endl;
  }
  iteration_++;
  return residual;
}


double BinaryBeliefPropagation::BeliefByGrid(string name, 
					     vector<int> &index) {
  BinaryVariable *variable = fg_->GetVariableByGrid(name, index);

  return variable->Belief();
}


double BinaryBeliefPropagation::Belief(BinaryVariable *variable) {
  return variable->Belief();
}
