// Copyright 2009 and onwards
// Author: dtarlow@cs.toronto.edu (Danny Tarlow)
// 7/25/09
//
// A binary factor graph object.  Its main job is to store
// the structure of relationships between nodes and factors, and
// to provide access to them.  After adding a factor or variable
// to the graph, the factor graph takes ownership of the memory.

#include <vector>
#include <map>
#include <utility>

#include "utils.h"
#include "binary_factor_graph.h"
#include "binary_variable.h"
#include "factor.h"
#include "edge.h"
#include "hyper_matrix.h"

BinaryFactorGraph::BinaryFactorGraph() { 
  variables_ = new vector<BinaryVariable *>();
  factors_ = new vector<Factor *>();
  singleton_factors_ = new vector<Factor *>();
  non_singleton_factors_ = new vector<Factor *>();
  edges_ = new vector<Edge *>();
  indices_ = new vector<pair<string, HyperMatrix<BinaryVariable *> *> >();
  layer_bounds_ = new vector<pair<string, vector<uint> *> >();
}


BinaryFactorGraph::~BinaryFactorGraph() { 
  delete variables_;
  delete factors_;
  delete edges_;
}


void BinaryFactorGraph::Print() {
  for (uint i = 0; i < variables_->size(); i++) {
    cout << variables_->at(i)->Name() << endl;
  }
}


void BinaryFactorGraph::AssignFromOtherFactorGraph(BinaryFactorGraph *other_fg) {
  const vector<BinaryVariable *> *other_variables = other_fg->Variables();
  
  CHECK_EQ(other_variables->size(), Variables()->size(), "Factor graphs of different sizes");
  
  for (uint i = 0; i < Variables()->size(); i++) {
    BinaryVariable *var = Variables()->at(i);
    BinaryVariable *other_var = other_variables->at(i);
    
    CHECK_EQ(var->Name(), other_var->Name(), "Variable orderings different");
    
    if (other_var->IsAssigned()) {
      var->AssignValue(other_var->Assignment());
    }
  }
}


void BinaryFactorGraph::AddVariableGrid(string name, const vector<uint> &dim) {
  HyperMatrix<BinaryVariable *> *variable_index = new HyperMatrix<BinaryVariable *>(dim);
  indices_->push_back(make_pair(name, variable_index));

  vector<uint> *bounds = new vector<uint>();
  for (uint i = 0; i < dim.size(); i++) {
    bounds->push_back(0);
  }
  layer_bounds_->push_back(make_pair(name, bounds));

  vector<int> res(dim.size());

  for (int i = 0; i < variable_index->SpaceNeeded(dim); i++) {
    string new_var_name = name;
    variable_index->AbsoluteToVectorIndex(i, &res);

    for (uint j = 0; j < dim.size(); j++) {
      new_var_name += StringPrintf("_%d", res[j]);
    }

    BinaryVariable *new_variable = new BinaryVariable(new_var_name);
    variable_index->SetAbsoluteIndex(i, new_variable);
    AddVariable(new_variable);
  }
}


void BinaryFactorGraph::AddBoundedVariableGrid(string name, 
					       const vector<uint> &bound,
					       const vector<uint> &dim) {
  HyperMatrix<BinaryVariable *> *variable_index = new HyperMatrix<BinaryVariable *>(dim);
  indices_->push_back(make_pair(name, variable_index));

  vector<uint> *bounds = new vector<uint>();
  for (uint i = 0; i < dim.size(); i++) {
    bounds->push_back(bound[i]);
  }
  layer_bounds_->push_back(make_pair(name, bounds));

  vector<int> res(dim.size());

  for (int i = 0; i < variable_index->SpaceNeeded(dim); i++) {
    string new_var_name = name;
    variable_index->AbsoluteToVectorIndex(i, &res);

    for (uint j = 0; j < dim.size(); j++) {
      new_var_name += StringPrintf("_%d", bound[j] + res[j]);
    }

    BinaryVariable *new_variable = new BinaryVariable(new_var_name);
    variable_index->SetAbsoluteIndex(i, new_variable);
    AddVariable(new_variable);
  }
}



HyperMatrix<BinaryVariable *> *BinaryFactorGraph::FullLayer(string layer) {
  for (uint i = 0; i < indices_->size(); i++) {
    if (indices_->at(i).first == layer) {
      return indices_->at(i).second;
    }
  }

  cerr << "Unknown layer: " << layer << endl;
  cerr << "Choices are ";
  for (uint i = 0; i < indices_->size(); i++) {
    cerr << indices_->at(i).first << " ";
  }
  cerr << endl;
  CHECK(false, "Should never reach here");
  return NULL;  
}


BinaryVariable *BinaryFactorGraph::GetVariableByGrid(string layer, 
						     const vector<int> &index) {

  vector<uint> *bounds = NULL;
  vector<int> query_index;
  for (uint i = 0; i < layer_bounds_->size(); i++) {
    if (layer_bounds_->at(i).first == layer) {
      bounds = layer_bounds_->at(i).second;
    }
  }
  if (bounds == NULL) {
    SHOW(layer);
    CHECK(false, "Didn't find layer");
  }

  for (uint i = 0; i < index.size(); i++) {
    CHECK_GE(index[i], static_cast<int>(bounds->at(i)), 
	     "There's no entry for this var in layer");
    query_index.push_back(index[i] - bounds->at(i));
  }

  return FullLayer(layer)->Get(query_index);

  /*
  for (uint i = 0; i < indices_->size(); i++) {
    if (indices_->at(i).first == name) {
      return indices_->at(i).second->Get(index);
    }
  }

  CHECK(false, "Should never reach here");
  return NULL;
  */
}


BinaryVariable *
BinaryFactorGraph::GetVariableByGridIfExists(string layer, 
					     const vector<uint> &bounds,
					     const vector<uint> &layer_size,
					     const vector<int> &index) {
  
  vector<int> query_index;
  for (uint i = 0; i < index.size(); i++) {
    if (index[i] >= static_cast<int>(bounds[i])  &&
	index[i] - static_cast<int>(bounds[i]) < static_cast<int>(layer_size[i])) {
      query_index.push_back(index[i] - bounds[i]);
    } else {
      return NULL;
    }    
  }

  return FullLayer(layer)->Get(query_index);

}



BinaryVariable *BinaryFactorGraph::GetVariableByGridIfExists(string layer, 
							     const vector<int> &index) {


  vector<uint> *bounds = NULL;
  vector<int> query_index;
  for (uint i = 0; i < layer_bounds_->size(); i++) {
    if (layer_bounds_->at(i).first == layer) {
      bounds = layer_bounds_->at(i).second;
    }
  }

  HyperMatrix<BinaryVariable *> *full_layer = FullLayer(layer);

  for (uint i = 0; i < index.size(); i++) {
    if (index[i] >= static_cast<int>(bounds->at(i))  &&
	index[i] - static_cast<int>(bounds->at(i)) < full_layer->SizeByDim(i)) {
      query_index.push_back(index[i] - bounds->at(i));
    } else {
      return NULL;
    }    
  }

  return full_layer->Get(query_index);

}


double BinaryFactorGraph::Likelihood() {
  double lik = 0;
  for (uint i = 0; i < Factors()->size(); i++) {
    Factor *factor = Factors()->at(i);

    lik += factor->Likelihood();
  }
  return lik;
}


void BinaryFactorGraph::LikelihoodByFactorClass(int_double_map &lik_by_type) {
  for (uint i = 0; i < factor_classes_.size(); i++) {
    lik_by_type[factor_classes_[i]] = 0;
  }

  for (uint i = 0; i < Factors()->size(); i++) {
    Factor *factor = Factors()->at(i);
    lik_by_type[factor->FactorClass()] += factor->Likelihood();
  }
}



double BinaryFactorGraph::LikelihoodByFeature(map<string, double> &lik_by_type) {
  double total_lik = 0;
  for (uint i = 0; i < Factors()->size(); i++) {
    Factor *factor = Factors()->at(i);

    map<string, double>::iterator iter = lik_by_type.find(factor->Feature());
    double factor_lik = factor->Likelihood();
    total_lik += factor_lik;
    if (iter != lik_by_type.end()) {
      lik_by_type[factor->Feature()] += factor_lik;
    } else {
      lik_by_type.insert(make_pair(factor->Feature(), factor_lik));
    }
  }
  return total_lik;
}



void BinaryFactorGraph::AddVariable(BinaryVariable *variable) {
  variables_->push_back(variable);
}


void BinaryFactorGraph::AddFactor(Factor *factor) {
  factor->SetDamping(damping_);

  // TODO: Should just use a set data structure here, but there
  // will only ever be a small number of classes, so this isn't
  // that wasteful (no internet to look up syntax).
  bool factor_class_exists = false;
  for (uint i = 0; i < factor_classes_.size(); i++) {
    if (factor->FactorClass() == factor_classes_[i]) {
      factor_class_exists = true;
      break;
    }
  }
  if (!factor_class_exists) {
    factor_classes_.push_back(factor->FactorClass());
  }

  // Add the factor and hook up edges between it and its scope.
  factors_->push_back(factor);
  
  for (uint i = 0; i < factor->Scope()->size(); i++) {
    AddEdge(factor->Scope()->at(i), factor);
  }
  
  if (factor->Scope()->size() == 1) {
    singleton_factors_->push_back(factor);
  } else {
    non_singleton_factors_->push_back(factor);
  }

}


void BinaryFactorGraph::AddEdge(BinaryVariable *variable, Factor *factor) {
  // Create a new edge and store it in the graph
  Edge *edge = new Edge(variable, factor);
  edges_->push_back(edge);

  // Let each of the endpoints know its connected to the edge.
  variable->AddEdge(edge);
  factor->AddEdge(edge);
}
