// Copyright 2006 and onwards
// Author: dtarlow@cs.toronto.edu (Danny Tarlow)
// 9/11/2006
//
// Interface defining graph structure functionality.

#ifndef CYBORG_BINARY_FACTOR_GRAPH_H__
#define CYBORG_BINARY_FACTOR_GRAPH_H__

using namespace std;

#include <vector>
#include <string>
#include <iostream>
#include <algorithm>
#include <map>

#include "hyper_matrix.h"

class Factor;
class Edge;
class BinaryVariable;
class Message;


typedef HyperMatrix<BinaryVariable*> VariableLayer;


class BinaryFactorGraph {
 public:
  BinaryFactorGraph();

  ~BinaryFactorGraph();

  void Print();

  // Graph modifiers for individual variables, variable grids, and factors
  void AddVariable(BinaryVariable *variable);

  void AddVariableGrid(string name, const vector<uint> &dim);

  void AddFactor(Factor *factor);


  // Graph modifiers for subsets of layers
  //
  // A variable grid where only variables lower_bound through lower_bound
  // + dim are active.  This is useful for subgraphs on patches of an image
  // without wastefully instantiating a variable for every pixel.  See the
  // superpixel model for an example.
  void AddBoundedVariableGrid(string name, 
			      const vector<uint> &lower_bound,
			      const vector<uint> &dim);

  BinaryVariable *GetVariableByGridIfExists(string name, 
					    const vector<uint> &bounds,
					    const vector<uint> &layer_size,
					    const vector<int> &dim);



  // Getters
  BinaryVariable *GetVariableByGrid(string name, 
				    const vector<int> &dim);
  BinaryVariable *GetVariableByGridIfExists(string name, 
					    const vector<int> &dim);
  const vector<Edge *> *Edges() { return edges_; }
  const vector<BinaryVariable *> *Variables() { return variables_; }
  const vector<Factor *> *Factors() { return factors_; }
  const vector<Factor *> *SingletonFactors() { return singleton_factors_; }
  const vector<Factor *> *NonSingletonFactors() { return non_singleton_factors_; }
  const vector<factor_class_t> *FactorClasses() { return &factor_classes_; }

  VariableLayer *FullLayer(string layer);


  // Dealing with assignments and likelihoods.  For learning, we need to
  // know where the contributions to the likelihood came from.
  double Likelihood();

  void LikelihoodByFactorClass(int_double_map &lik_by_class);

  double LikelihoodByFeature(map<string, double> &lik_by_feature);

  double AssignVariables(vector<bool> &assignment);

  void AssignFromOtherFactorGraph(BinaryFactorGraph *other_fg);

  // Scheduling and algorithm execution parameters

  void SetDamping(double damping) { damping_ = damping; }

 protected:

  // This is called automatically for each factor's scope when
  // the factor is added to the factor graph.
  void AddEdge(BinaryVariable *variable, Factor *factor);

  vector<BinaryVariable *> *variables_;
  vector<Factor *> *factors_;
  vector<Factor *> *singleton_factors_;
  vector<Factor *> *non_singleton_factors_;
  vector<Edge *> *edges_;

  vector<pair<string, VariableLayer *> > *indices_;

  vector<factor_class_t> factor_classes_;

  // Allow layers to not range over the full image size, for example
  vector<pair<string, vector<uint> *> > *layer_bounds_;

  double damping_;

};

#endif
