#include "binary_variable.h"
#include "definitions.h"
#include <string>
#include <cmath>
#include "edge.h"
#include "factor.h"

BinaryVariable::BinaryVariable(string name) {
  name_ = name;
  edges_ = new vector<Edge *>();
  outgoing_messages_ = new vector<double>();
  outgoing_sum_ = 0;
  is_dirty_ = true;
  damping_ = DEFAULT_DAMPING;
}


BinaryVariable::~BinaryVariable() {
  delete outgoing_messages_;
  delete edges_;
}

void BinaryVariable::AddEdge(Edge *edge) {
  edge->SetPositionInVariable(edges_->size());
  edges_->push_back(edge);
  outgoing_messages_->push_back(0);
}


bool BinaryVariable::IsDirty() {
  return is_dirty_;
}


void BinaryVariable::Touch() {
  is_dirty_ = true;
}


void BinaryVariable::Touch(int position_in_variable) {
  Touch();
}


void BinaryVariable::ComputeOutgoingMessage(int position_in_variable) {
  outgoing_sum_ = 0;
  for (uint i = 0; i < edges_->size(); i++) {
    outgoing_sum_ += edges_->at(i)->ToVariable();
  }

  outgoing_messages_->at(position_in_variable) = 
    outgoing_sum_ - edges_->at(position_in_variable)->ToVariable();

}


void BinaryVariable::ComputeAllOutgoingMessages() {
  outgoing_sum_ = 0;
  for (uint i = 0; i < edges_->size(); i++) {
    outgoing_sum_ += edges_->at(i)->ToVariable();
    //cout << edges_->at(i)->GetFactor()->Name() << " --> " << Name() << ": " 
    //<< edges_->at(i)->ToVariable() << endl;

  }
  
  for (uint i = 0; i < outgoing_messages_->size(); i++) {
    outgoing_messages_->at(i) = outgoing_sum_ - edges_->at(i)->ToVariable();
  }
  
  is_dirty_ = false;

}



void BinaryVariable::ComputeAllOutgoingMessagesExcept(int pos) {
  outgoing_sum_ = 0;
  for (uint i = 0; i < edges_->size(); i++) {
    outgoing_sum_ += edges_->at(i)->ToVariable();
    //cout << edges_->at(i)->GetFactor()->Name() << " --> " << Name() << ": " 
    //<< edges_->at(i)->ToVariable() << endl;

  }
  
  for (uint i = 0; i < outgoing_messages_->size(); i++) {
    if (static_cast<int>(i) != pos) {
      outgoing_messages_->at(i) = outgoing_sum_ - edges_->at(i)->ToVariable();
    }
  }
  
  is_dirty_ = false;

}



double BinaryVariable::UpdateOutgoingMessage(int position_in_variable) {
  // Don't damp variable --> factor messages
  double old_message = edges_->at(position_in_variable)->ToFactor();
  double new_message = outgoing_messages_->at(position_in_variable);

  string factor_name = edges_->at(position_in_variable)->GetFactor()->Name();
  CHECK_SMALL(old_message, "Old message huge val " + factor_name);
  CHECK_SMALL(new_message, "New message huge val " + factor_name);

  double residual = abs(new_message - old_message);
  edges_->at(position_in_variable)->SetToFactor(new_message);
  return residual;
}


void BinaryVariable::Neighbors(vector<Factor *> &neighbors) {
  for (uint i = 0; i < edges_->size(); i++) {
    neighbors.push_back(edges_->at(i)->GetFactor());
  }
}


/*
void BinaryVariable::UpdateMessage(int position_in_variable, double value) {
  edges_->at(position_in_variable)->SetToFactor(value);
}
*/

double BinaryVariable::UpdateAllOutgoingMessages() {
  double residual = 0;
  for (uint i = 0; i < edges_->size(); i++) {
    residual += UpdateOutgoingMessage(i);
  }
  return residual;
}



double BinaryVariable::UpdateAllOutgoingMessagesExcept(int pos) {
  double residual = 0;
  for (uint i = 0; i < edges_->size(); i++) {
    if (static_cast<int>(i) != pos) {
      residual += UpdateOutgoingMessage(i);
    }
  }
  return residual;
}


double BinaryVariable::Belief() {
  double belief = 0;
  for (uint i = 0; i < edges_->size(); i++) {
    belief += edges_->at(i)->ToVariable();
  }

  return belief;
}
