#include "counts_factor.h"

#include <iostream>
#include <stdio.h>
#include "utils.h"

#include "binary_variable.h"

using namespace std;


CountsFactor::CountsFactor(string name) : Factor(name) {
  N_ = 0;

  lr_ = NULL;
  rl_ = NULL;

  potential_ = NULL;
}

CountsFactor::~CountsFactor() {
  delete incoming_messages_;

  delete lr_;
  delete rl_;

}


void CountsFactor::LoadIncomingMessagesFromEdges() {
  CHECK_EQ(edges_->size(), incoming_messages_->size(),
	   "Edges and incoming messages of different size");
  for (uint i = 0; i < edges_->size(); i++) {
    SetIncomingMessage(i, edges_->at(i)->ToFactor());
  }
}


void CountsFactor::SetIncomingMessage(int var, double val) {
  // Check to see if we're changing the message value before we make the
  // factor dirty.  Later, we might want to check to see if the difference 
  // is within some tolerance.
  if (val != incoming_messages_->at(var)) {
    incoming_messages_->at(var) = val;
    Touch();
  }
}


void CountsFactor::SetPotential(vector<double> &f) {
  if (potential_ == NULL) {
    potential_ = new vector<double>(f.size());
  } else if (potential_->size() != f.size()) {
    delete potential_;
    potential_ = new vector<double>(f.size());
  }

  for (uint i = 0; i < f.size(); i++) {
    potential_->at(i) = f[i];
  }
}


void CountsFactor::ComputeOutgoingMessage(int position_in_factor) {
  //EXPECT(false, "Inefficient call to CountsFactor::ComputeOutgoingMessage(int)");

  ComputeAllOutgoingMessages();
}


void CountsFactor::ComputeAllOutgoingMessages() {
  LoadIncomingMessagesFromEdges();
  N_ = static_cast<int>(edges_->size());

  // Sort incoming messages while keeping track of the original
  // indices.
  vector<pair<double, int> > vals_and_indices(N_);
  for (int i = 0; i < N_; i++) {
    vals_and_indices[i] = make_pair(incoming_messages_->at(i), i);
  }
  sort(vals_and_indices.begin(), vals_and_indices.end(), PairCompare);

  // Keep a reverse index so we can look up where an original index
  // ended up.
  vector<int> ridx(N_);
  for (int i = 0; i < N_; i++) {
    ridx[vals_and_indices[i].second] = i;
  }

  // We need to align cumulative sums with the #vars potential function
  // in several ways.  This probably doesn't make a ton of sense unless
  // you've worked through the derivations.
  vector<double> cum_ws(N_ + 1);
  vector<double> cum_w_s_1(N_ + 1);
  vector<double> cum_w_s0(N_ + 1);
  vector<double> cum_w_s1(N_ + 1);
  for (int i = 0; i <= N_; i++) {
    // Cumulative sums of sorted messages
    if (i == 0) {
        cum_ws[i] = 0; //potential_->at(0);
    } else {
      cum_ws[i] = vals_and_indices[i - 1].first + cum_ws[i - 1];
    }

    // Align cum_ws exactly with f
    cum_w_s0[i] = cum_ws[i] + potential_->at(i);

    // Shift f one to the left.
    if (i > 0) {
      cum_w_s_1[i] = cum_ws[i] + potential_->at(i - 1);

    } else {
      cum_w_s_1[i] = -BIG_DOUBLE;
    }

    // Shift f one to the right.
    if (i < N_) { // - 1) {
      cum_w_s1[i] = cum_ws[i] + potential_->at(i + 1);
    } else {
      cum_w_s1[i] = -BIG_DOUBLE;
    }
  }

  // Cumulative maxes.  We need to keep cumulative maxes from both directions
  // for the zero-offset alignment.  These are stored in m0R and m0L.
  vector<double> m_1(N_+1);
  vector<double> m0R(N_+1);
  vector<double> m0L(N_+1);
  vector<double> m1(N_+1);

  for (int i = 0; i <= N_; i++) {
    if (i == 0) {
      m1[i] = cum_w_s1[i];
      m0L[i] = cum_w_s0[i];
      m0R[N_-i] = cum_w_s0[N_-i];
      m_1[N_-i] = cum_w_s_1[N_-i];
    } else {
      m1[i] = MAX(cum_w_s1[i], m1[i-1]);
      m0L[i] = MAX(cum_w_s0[i], m0L[i-1]);
      m0R[N_-i] = MAX(cum_w_s0[N_-i], m0R[N_-i+1]);
      m_1[N_-i] = MAX(cum_w_s_1[N_-i], m_1[N_-i+1]);
    }
  }

  int pos;
  double msg0;
  double msg1;
  for (int i = 0; i < N_; i++) {
    pos = ridx[i];

    msg0 = -HUGE_VAL;
    msg1 = -HUGE_VAL;

    if (pos > 0) {
        msg0 = MAX(msg0, m0L[pos - 1]);
        msg1 = MAX(msg1, m1[pos - 1]);
    }
    if (pos < N_) {
        msg0 = MAX(msg0, m_1[pos + 1] - vals_and_indices[pos].first);
        msg1 = MAX(msg1, m0R[pos + 1] - vals_and_indices[pos].first);        
    }
    
    outgoing_messages_->at(i) = msg1 - msg0;
  }
}


double CountsFactor::Likelihood() {
  int num_on = 0;
  for (uint i = 0; i < scope_->size(); i++) {
    BinaryVariable *var = scope_->at(i);
    CHECK(var->IsAssigned(), "Variable not assigned");
    if (var->Assignment()) {
      ++num_on;
    }
  }
  return potential_->at(num_on);
}



/*
void CountsFactor::UpdateAllOutgoingMessages() {
  CHECK_EQ(outgoing_messages_->size(), edges_->size(),
	   "Edges_ and outgoing_messages_ have different sizes");

  for (uint i = 0; i < outgoing_messages_->size(); i++) {
    edges_->at(i)->SetToVariable(outgoing_messages_->at(i));
  }
  }
*/



/*
int test_sequence_factor(char *argv, int argc) {

  int N = 1000;
  CountsFactor *seq = new CountsFactor(N);


  double weights[N];
  double forced_zeros[N];
  double forced_ones[N];

  for (int i = 0; i < N; i++) {
    double x = i / 2.0 - 20.0;
    weights[i] = x * x / 100 - 5 * (i - 50) + 10;
  }

  cout << "Weights:" << endl;
  for (int i = 0; i < N; i++) {
    cout << weights[i] << " ";
  }
  cout << endl << endl;

  seq->set_incoming_messages(weights);

  seq->compute_all_messages();

  for (int i = 0; i < N; i++) {
    cout << seq->message(i) << " ";
  }
  cout << endl;

}
*/
