// Copyright 2009 and onwards
// Author: dtarlow@cs.toronto.edu (Danny Tarlow)
// 7/20/2009
//
// Main executable for the CYBORG probabilistic reasoner.

#include <iostream>
#include <string>
#include <vector>
#include <algorithm>
#include <ext/hash_set>
#include <boost/program_options.hpp>
#include <cmath>
#include <time.h>
#include <map>
#include <utility>

#include "utils.h"
#include "definitions.h"
#include "binary_factor_graph.h"
#include "binary_variable.h"
#include "binary_belief_propagation.h"
#include "factor.h"
#include "singleton_factor.h"
#include "potts_factor.h"
#include "counts_factor.h"
#include "sequence_factor.h"

using namespace std;
namespace po = boost::program_options;

double RandomPotential(double scale) {
  // Return a potential in range [-scale, scale]
  double rand_01 = rand() / static_cast<double>(RAND_MAX);
  return scale * (2 * rand_01 - 1);
}


double RandomPositivePotential(double scale) {
  // Return a potential in range [-scale, scale]
  return scale * (rand() / static_cast<double>(RAND_MAX));
}


int RandomColor(int layer, int channel) {
  return 255 * (rand() / static_cast<float>(RAND_MAX));
  /*
  int prime1 = 7919;
  int prime2 = 30029;

  return (layer * prime1 + (1 + channel) * prime2) % 255;
  */
}


int main(int argc, char **argv) {

  po::options_description options("Allowed options");

  int num_iterations;
  double damping;
  int idim;
  int jdim;
  bool use_counts;
  double counts_strength;
  bool use_convexity;
  bool use_pairwise;
  double pairwise_strength;
  double fraction_on;
  message_schedule_t schedule = SYNCHRONOUS;
  int schedule_int;
  string codename = "";
  string intermediate_results_filename ="";

  int seed;

  options.add_options()
    ("help", "produce help message")
    ("idim", 
     po::value<int>(&idim)->default_value(20),
     "size in i dimension")
    ("jdim", 
     po::value<int>(&jdim)->default_value(180),
     "size in j dimension")
    ("damping", 
     po::value<double>(&damping)->default_value(DEFAULT_DAMPING),
     "damping factor.  higher damping means use less of new message.")
    ("iterations", 
     po::value<int>(&num_iterations)->default_value(250),
     "maximum number of iterations")
    ("seed", 
     po::value<int>(&seed)->default_value(1),
     "random seed. use -1 for srand(time(NULL))")
    ("schedule",
     po::value<int>(&schedule_int)->default_value(0),
     "message schedule: 0 - SYNCHRONOUS")
    ("use_counts", 
     po::value<bool>(&use_counts)->default_value(true),
     "use a counts factor in the model")
    ("counts_strength",
     po::value<double>(&counts_strength)->default_value(1.0),
     "how much to scale penalty for deviating from desired count")
    ("fraction_on",
     po::value<double>(&fraction_on)->default_value(.75),
     "expected fraction of pixels that are 'on'")
    ("use_convexity", 
     po::value<bool>(&use_convexity)->default_value(false),
     "use a convexity factor in the model")
    ("use_pairwise", 
     po::value<bool>(&use_pairwise)->default_value(true),
     "use pairwise factors between horizontally connected variables")
    ("pairwise_strength", 
     po::value<double>(&pairwise_strength)->default_value(1.0),
     "strength of potential between neighbors (positive, higher is stronger)");
  
  po::positional_options_description positional_options;

  po::variables_map vm;
  po::store(po::command_line_parser(argc, argv).
	    options(options).positional(positional_options).run(), vm);
  po::notify(vm);


  if (seed == -1) {
    srand(time(NULL));
  } else {
    srand(seed);
  }
  
  // Is there a better way to handle enums with boost::program_options?
  if (schedule_int == 0) {
    schedule = SYNCHRONOUS;
  } else {
    SHOW(schedule_int);
    EXPECT(false, "Unknown schedule_int");
  }

  if (vm.count("help")) {
    cerr << options << endl;
    return 1;
  }

  BinaryFactorGraph *fg = new BinaryFactorGraph();
  fg->SetDamping(damping);
  vector<uint> dim(2); 
  dim[0] = idim; 
  dim[1] = jdim; 
  vector<int> index(2); 
  fg->AddVariableGrid("foreground", dim);


  if (use_counts && counts_strength != 0) {
    CountsFactor *counts_factor = new CountsFactor("counts_factor_1"); 
    // Scope
    for (int i = 0; i < idim; i++) { 
      for (int j = 0; j < jdim; j++) { 
	index[0] = i; 
	index[1] = j; 
	BinaryVariable *var = fg->GetVariableByGrid("foreground", index); 
	counts_factor->AddToScope(var); 
      } 
    } 
    // Potential
    vector<double> potential(idim * jdim + 1); 
    for (int i = 0; i <= idim * jdim; i++) {
      double diff = i - fraction_on * idim * jdim; 
      potential[i] = -counts_strength * diff * diff;
    } 
    counts_factor->SetPotential(potential);
    fg->AddFactor(counts_factor); 
  }

  if (use_convexity) {
    for (int j = 0; j < jdim; j++) {
      SequenceFactor *sequence_factor = 
	new SequenceFactor(StringPrintf("sequencev_%d", j));
      for (int i = 0; i < idim; i++) {
	index[0] = i;
	index[1] = j;
	BinaryVariable *var = fg->GetVariableByGrid("foreground", index);
	sequence_factor->AddToScope(var);
      }
      fg->AddFactor(sequence_factor);
    }    
  }

  // Singleton potentials
  for (int i = 0; i < idim; i++) { 
    for (int j = 0; j < jdim; j++) { 
      string factor_name = StringPrintf("s_%d_%d", i, j); 
      SingletonFactor *s_ij = new SingletonFactor(factor_name); 
      index[0] = i; index[1] = j; 
      BinaryVariable *var = fg->GetVariableByGrid("foreground", index); 
      s_ij->AddToScope(var); 
      s_ij->SetPotential(RandomPotential(1.0));
      fg->AddFactor(s_ij); 
    } 
  } 

  if (use_pairwise && pairwise_strength != 0) {
    // Horizontal pairwise potentials.  You probably don't usually
    // want only horizontal potentials, but this is for illustration
    // purposes and makes it easier to visualize the effect of pairwise
    // terms.
    for (int i = 0; i < idim; i++) {
      for (int j = 0; j < jdim - 1; j++) {
	index[0] = i; index[1] = j;
	BinaryVariable *var_L = fg->GetVariableByGrid("foreground", index);
	index[0] = i; index[1] = j + 1;
	BinaryVariable *var_R = fg->GetVariableByGrid("foreground", index);

	string factor_name = StringPrintf("h_%d_%d_%d", i, j, j + 1);
	PottsFactor *theta_LR = new PottsFactor(factor_name);

	theta_LR->AddToScope(var_L);
	theta_LR->AddToScope(var_R);
	
	theta_LR->SetPotential(RandomPositivePotential(pairwise_strength));

	fg->AddFactor(theta_LR);
      }
    }
  }

  BinaryBeliefPropagation *bp = new BinaryBeliefPropagation(fg);
  cout << endl;
  bool converged = false;
  double residual = -1;
  for (int i = 0; i < num_iterations; i++) {
    residual = bp->PassMessages();

    bp->AssignVariablesByBelief();  // Assign just by belief

    cout << "Iteration " << i << "\t\tLikelihood = " << fg->Likelihood() 
	 << "\t\tResidual=" << residual;
    for (int j = 0; j < 10; j++) { cout << " "; }

    if (residual <= 1e-5) {
      cout << endl << "CONVERGED!" << endl;
      converged = true;
      break;
    } else {
      cout << "\r";
      cout.flush();
    }
  }
  cout << endl;

  int num_on = 0;
  int total_vars = 0;
  for (int i = 0; i < idim; i++) {
    for (int j = 0; j < jdim; j++) {
      index[0] = i;
      index[1] = j;
      bool var_on = fg->GetVariableByGrid("foreground", index)->Assignment();
      if (var_on) {
	cout << "X";
	num_on++;
      } else {
	cout << ".";
      }
      total_vars++;
    }
    cout << endl;
  }
  cout << num_on << " / " << total_vars << " (" 
       << 100.0 *num_on / static_cast<float>(total_vars) << "%) variables on" << endl;
  cout << endl;
  
  cout << "Likelihood = " << fg->Likelihood() << "\t\tResidual=" << residual << endl;

  if (!converged) {
    cout << "DID NOT CONVERGE" << endl;
    cout << "Results may still be reasonable, but try increasing --iterations or changing --damping"
	 << endl << "(or decreasing intensity of potentials)" << endl;
  } else {
    cout << "CONVERGED" << endl;
  }

}
