#include "factor.h"

#include <cmath>
#include "binary_variable.h"

Factor::Factor(string name) {
  name_ = name;
  x_pos_ = -1;
  y_pos_ = -1;
  edges_ = new vector<Edge *>();
  scope_ = new vector<BinaryVariable *>();
  outgoing_messages_ = new vector<double>();
  incoming_messages_ = new vector<double>();
  is_dirty_ = true;
  damping_ = DEFAULT_DAMPING;
  test_mode_ = false;
  positive_incoming_message_sum_ = 0;
  negative_incoming_message_sum_ = 0;
  num_updates_ = 0;
}


void Factor::AddEdge(Edge *edge) {
  int position_in_factor = edges_->size();
  edge->SetPositionInFactor(position_in_factor);
  
  edges_->push_back(edge);
  outgoing_messages_->push_back(0);
  incoming_messages_->push_back(0);
  
  is_dirty_vector_.push_back(true);
  is_dirty_ = true;
}


double Factor::UpdateOutgoingMessage(int position_in_factor) {
  CHECK_GE(position_in_factor, 0, "Can't compute an out of bounds message");
  CHECK_LT(position_in_factor, static_cast<int>(outgoing_messages_->size()), 
	    "Can't compute an out of bounds message");

  // Damp factor --> variable messages
  double old_message = edges_->at(position_in_factor)->ToVariable();
  double new_message = outgoing_messages_->at(position_in_factor);
  double damped_value = DampedMessageValue(new_message,
					   old_message,
					   damping_);
  double residual = abs(new_message - old_message);

  edges_->at(position_in_factor)->SetToVariable(damped_value);
  edges_->at(position_in_factor)->GetVariable()->Touch();
  num_updates_++;

  return residual;
}



void Factor::ComputeAllOutgoingMessagesExcept(int except_pos) {
  CHECK_EQ(outgoing_messages_->size(), edges_->size(),
	   "Edges_ and outgoing_messages_ have different sizes");

  for (int i = 0; i < static_cast<int>(outgoing_messages_->size()); i++) {
    if (i != except_pos) {
      ComputeOutgoingMessage(i);
    }
  }
}


double Factor::UpdateAllOutgoingMessages() {
  CHECK_EQ(outgoing_messages_->size(), edges_->size(),
	   "Edges_ and outgoing_messages_ have different sizes");
  double residual = 0;
  for (int i = 0; i < static_cast<int>(outgoing_messages_->size()); i++) {
    residual += UpdateOutgoingMessage(i);
  }
  return residual;
}


void Factor::UpdateIncomingMessage(int position_in_edges, 
				   double old_message, 
				   double new_message) {

  if (new_message == old_message) {
    return;
  }

  incoming_messages_->at(position_in_edges) = new_message;

  if (old_message > 0) {
    positive_incoming_message_sum_ -= old_message;
  } else {
    negative_incoming_message_sum_ -= old_message;
  }
  
  if (new_message > 0) {
    positive_incoming_message_sum_ += new_message;
  } else {
    negative_incoming_message_sum_ += new_message;
  }  
}

double Factor::DampedMessageValue(double new_message,
				  double old_message,
				  double damping) {
  return damping * old_message + (1.0 - damping) * new_message;
}
