// copyright 2009 and onwards
// Author: dtarlow@cs.toronto.edu (Danny Tarlow)
// 7/25/09

#ifndef CYBORG_FACTOR_H__
#define CYBORG_FACTOR_H__

using namespace std;

#include <vector>
#include <string>
#include <iostream>
#include "definitions.h"
#include "edge.h"

class Factor {

 public:
  Factor(string name);

  virtual ~Factor() {
    delete edges_;
    delete scope_;
    delete outgoing_messages_;
    delete incoming_messages_;

  }

  virtual void SetTestMode(bool test_mode) {
    test_mode_ = test_mode;
  }

  void SetPosition(double x, double y) { 
    x_pos_ = x;
    y_pos_ = y;
  }

  double XPosition() { return x_pos_; }

  double YPosition() { return y_pos_; }

  void SetDamping(double damping) { damping_ = damping; }


  virtual void ComputeAllOutgoingMessages() = 0;

  virtual void ComputeAllOutgoingMessagesExcept(int except_pos);

  virtual void ComputeOutgoingMessage(int position_in_factor) = 0;

  virtual double UpdateOutgoingMessage(int position_in_factor);

  virtual double UpdateAllOutgoingMessages();
  
  virtual double Likelihood() {
    CHECK(false, "Likelihood fell back to base class implementation");
  }

  //virtual double Likelihood() = 0;

  double OutgoingMessage(int position_in_factor) {
    return outgoing_messages_->at(position_in_factor);
  }

  virtual void AddEdge(Edge *edge);

  virtual void Print() {
    cout << Name();
  }

  // Keep a running total of sums of positive and negative incoming messages
  // separately.  Optionally, also keep a dynamically sorted heap of incoming
  // messages.  These will be useful in computing outgoing messages in many
  // cases.
  virtual void UpdateIncomingMessage(int position_in_edges, 
				     double old_message, 
				     double new_message);

  inline virtual void AddToScope(BinaryVariable *variable) {
    scope_->push_back(variable);
  }

  inline string Name() { return name_; }

  inline vector<BinaryVariable *> *Scope() { return scope_; }

  inline virtual void Touch() { is_dirty_ = true; }

  inline virtual void Touch(int position_in_factor) {
    is_dirty_vector_[position_in_factor] = true;
  }

  inline virtual void Clean() { is_dirty_ = false; }

  inline virtual bool IsDirty() { return is_dirty_; }
  
  const vector<Edge *> *Edges() { return edges_; }

  int NumUpdates() { return num_updates_; }

  inline void SetFactorClass(factor_class_t factor_class) {
    factor_class_ = factor_class;
  }

  inline factor_class_t FactorClass() {
    return factor_class_;
  }

  inline void SetFeature(string feature) {
    feature_ = feature;
  }

  inline string Feature() {
    return feature_;
  }

 protected:
  double DampedMessageValue(double new_message, 
			    double old_message, 
			    double damping);

  string name_;
  bool is_dirty_;
  vector<bool> is_dirty_vector_;

  double damping_;
  bool test_mode_;
  double positive_incoming_message_sum_;
  double negative_incoming_message_sum_;

  bool use_dynamic_heap_;
  
  bool is_in_factor_graph_;

  factor_class_t factor_class_;  // Useful for learning, to group sufficient statistics

  string feature_;  // A more general form of factor_class that can be an arbitrary string


  vector<Edge *> *edges_;
  vector<BinaryVariable *> *scope_;

  // These are intermediate storage.  Real messages exist on edges once they are
  // actually passed.
  vector<double> *outgoing_messages_;
  vector<double> *incoming_messages_;

  // These are only used for coming up with schedules.  It is useful
  // to think of a factor between x_i and x_{i+1} as having x_pos of
  // i + .5.
  double x_pos_;
  double y_pos_; 

  int num_updates_;  // number of times update messages was called
};

#endif
