// Copyright 2006 and onwards
// Author: dtarlow@cs.toronto.edu (Danny Tarlow)
// 9/16/2006
//
// Templated definitions of simple hyper matrix class that allows
// for getting and setting elements in an arbitrary dimensional
// array.

template <class Element>
HyperMatrix<Element>::HyperMatrix() {
  elements_ = new vector<Element>();
}

template <class Element>
HyperMatrix<Element>::HyperMatrix(const vector<uint> &cardinalities) {
  int space_needed = SpaceNeeded(cardinalities);

  // Copy construct cardinalities.
  cardinalities_ = cardinalities;

  // Allocate an array with space for all elements.
  elements_ = new vector<Element>(space_needed);
}

template <class Element>
HyperMatrix<Element>::~HyperMatrix() {
  delete elements_;
}

template <class Element>
int HyperMatrix<Element>::SpaceNeeded(const vector<uint> &cardinalities) 
  const{
  int space_needed;

  // Calculate how much space is needed to store all entries.
  if (cardinalities.empty()) {
    space_needed = 0;
  } else {
    space_needed = 1;
    for (uint i = 0; i < cardinalities.size(); i++) {
      if (cardinalities[i] != 0) {
	space_needed *= cardinalities[i];
      }
    }  
  }
  return space_needed;
}

template <class Element>
void HyperMatrix<Element>::Resize(const vector<uint> &cardinalities) {

  CHECK_NOT_NULL(elements_, "resizing before elements_ has been allocated");

  cardinalities_ = cardinalities;

  uint space_needed = static_cast<uint>(SpaceNeeded(cardinalities));
  if (space_needed > elements_->size()) {
    elements_->resize(space_needed);
  } else if (space_needed < elements_->size()) {
    elements_->erase(elements_->begin() + space_needed, elements_->end());
  }
}

template <class Element>
void HyperMatrix<Element>::Set(const vector<int> &index, Element elem) {
  (*elements_)[AbsoluteIndex(index)] = elem;
}

template <class Element>
void HyperMatrix<Element>::SetAbsoluteIndex(int absolute_index, Element elem) {
  (*elements_)[absolute_index] = elem;
}


template <class Element>
void HyperMatrix<Element>::SetConstantValue(Element elem) {
  for (uint i = 0; i < elements_->size(); i++) {
    elements_->at(i) = elem;
  }  
}


template <class Element>
void HyperMatrix<Element>::LoadFromFile(string filename) {
  ifstream file(filename.c_str());
  string line;  

  CHECK(file.is_open(), "Error reading file " + filename);

  // Read dimensions from first line
  getline(file, line);
  char_separator<char> space_sep(" ");
  tokenizer<char_separator<char> > first_line_tokens(line, space_sep);
  tokenizer<char_separator<char> >::iterator beg;
  
  vector<uint> dim;
  for (beg=first_line_tokens.begin(); beg != first_line_tokens.end(); ++beg) {
    dim.push_back(static_cast<uint>(atoi((*beg).c_str())));
  }
  cout << "Dim: " << dim[0] << "," << dim[1] << endl;

  // Allocate the hyper matrix
  Resize(dim);

  if (dim.size() == 2) {
    // If it's a 2D file, we can read it as a dense matrix.
    vector<int> index(2);
    index[0] = 0;

    // Read lines
    while (getline(file, line)) {
      index[1] = 0;
      
      tokenizer<char_separator<char> > tokens(line, space_sep);
      for (beg = tokens.begin(); beg != tokens.end(); ++beg) {
	// Try to cast from a double to the type of element that we're using.
	// Is there a better way to do this?
	Element val = static_cast<Element>(atof((*beg).c_str()));
	Set(index, val);
	index[1]++;
      }

      index[0]++;
    }   
  }
}


template <class Element>
Element HyperMatrix<Element>::Get(const vector<int> &index) const {
  return (*elements_)[AbsoluteIndex(index)];
}


template <class Element>
Element HyperMatrix<Element>::GetAbsoluteIndex(int absolute_index) const {
  return elements_->at(absolute_index);
}


template <class Element>
void HyperMatrix<Element>::AbsoluteToVectorIndex(int absolute_index, 
						 vector<int> *vector_index) {
  int base = 1;
  for (uint i = 0; i < cardinalities_.size(); i++) {
    base *= cardinalities_[i];
    vector_index->at(i) = 0;
  }
  CHECK_NEQ(base, 0, "Base cannot be zero");

  int current_index = absolute_index;
  int digit;
  for (int i = static_cast<int>(cardinalities_.size()) - 1; i >= 0; i--) {
    base /= cardinalities_[i];
    digit = current_index / base;
    current_index -= digit * base;
    vector_index->at(i) = digit;
  }

  CHECK_EQ(AbsoluteIndex(*vector_index), absolute_index,
	   "AbsoluteToVectorIndex and AbsoluteIndex are inconsistent");
}


template <class Element>
int HyperMatrix<Element>::AbsoluteIndex(const vector<int> &index) const {
  uint absolute_index = 0;
  int base = 1;
  for (uint i = 0; i < index.size(); i++) {
    // Don't allow overstepping indices
    CHECK_LT(static_cast<uint>(index[i]), cardinalities_[i],
	     "hypermatrix index " + StringPrintf("%d", i) + " out of bounds");

    // Increment absolute_index appropriately.
    absolute_index += index[i] * base;
    
    // Move base to represent one digit over.
    base *= cardinalities_[i];
  }
  return absolute_index;
}

template <class Element>
void HyperMatrix<Element>::Print() {
  for (uint i = 0; i < elements_->size(); i++) {
    cout << (*elements_)[i] << " ";
    if (i % 10 == 9) cout << endl;      
  }
  cout << endl;
}
