#include "message_queue.h"
#include "message.h"

#include "pos_heap.h"
#include "heapplus.h"
#include "factor.h"
#include "binary_variable.h"


MessageQueue::MessageQueue(message_schedule_t schedule) { 
  if (schedule == RESIDUAL) {
    message_heap_ = new vector<Message *>();
    message_heap_nodes_ = new vector<Message *>();

  } else if (schedule == ASYNCHRONOUS) {
    message_heap_ = new vector<Message *>();
    message_heap_nodes_ = new vector<Message *>();
    asynch_pos_in_heap_ = 0;
  }

  debug_ = false;
  schedule_ = schedule;
  messages_sent_ = 0;
}


void MessageQueue::AddMessage(Message *message, bool new_message) {
  //cout << "Adding " << *message << endl;
  // Add to and update heap
  if (schedule_ == RESIDUAL) {
    double message_difference = 
      message->NewValue() - message->CurrentValue();
    if (message_difference < 0) message_difference *= -1;
    message->SetPriority(message_difference);

    message_heap_->push_back(message);
    push_heap_pos(message_heap_->begin(), message_heap_->end(), CompareHeapNodeStar());

  } else if (schedule_ == ASYNCHRONOUS) {
    message_heap_->push_back(message);
  }

  if (new_message) {
    // Add to the stable stored messages list, so that we can look up messages
    // based on their initial position.
    message->SetID(message_heap_nodes_->size());
    message_heap_nodes_->push_back(message);
  }
}


Message *MessageQueue::TopMessage() {
  if (schedule_ == ASYNCHRONOUS) {
    return message_stack_->top();
  } else if (schedule_ == RESIDUAL) {
    return message_heap_->at(0);
  } else {
    SHOW(schedule_);
    CHECK(false, "Unknown schedule");
    return NULL;
  }
}


void MessageQueue::UpdateMessagePriority(Message *message) {
  if (schedule_ == RESIDUAL) {
    // Actually change the message priority.  Assume message->ComputeMessage() has already
    // been called.
    double message_difference = 
      message->NewValue() - message->CurrentValue();
    
    if (message_difference < 0) message_difference *= -1;
    
    if (debug_) {
      cout << "     Set priority of " << *message << " to "
	   << message_difference << endl;
    }
    message->SetPriority(message_difference);
    UpdateHeapAtPosition(message->HeapPosition());      
  } else {
    CHECK(false, "Shouldn't be messing with message priorities if not RBP");
  }

  /*
  if (message->Priority() == 0 && message->IsActive()) {
    SetMessagePriorityToZero(message);
  } else if (message->Priority() == 0 && !message->IsActive()) {
    // Do nothing
  } else if (!message->IsActive() && message->Priority() > 0) {
    AddMessage(message, false);
  } else if (message->Priority() > 0) {
    CHECK_GE(message->HeapPosition(), 0, "Updating heap for a message with negative position");
    UpdateHeapAtPosition(message->HeapPosition());    
  } else {
    CHECK(false, "Unexpected case");
  }
  */

}


// To be called after pushing a message update to the edge.  There is no difference
// in the message, so set it to be zero.
void MessageQueue::SetMessagePriorityToZero(Message *message) {
  CHECK(false, "DEPRECATED");
  message->SetPriority(0);
  UpdateHeapAtPosition(message->HeapPosition());
  //message_heap_->pop_back();
  //message->SetPriority(0);
  //message->Disable();
}


void MessageQueue::UpdateHeapAtPosition(int position_in_heap) {
  if (schedule_ == RESIDUAL) {
    update_heap_pos(message_heap_->begin(), message_heap_->end(),
		    message_heap_->begin() + position_in_heap,
		    CompareHeapNodeStar());
  } else {
    CHECK(false, "Shouldn't be messing with heap position if not RBP");
  }
}


void MessageQueue::StartIteration() {
  if (schedule_ == RESIDUAL) {
    cout << "Unique messages updated: " << messages_passed_.size();
    messages_passed_.clear();
  }
  cout << " (" << messages_passed_.size() << ")" << endl;
}


bool MessageQueue::PassMessage() {
  messages_sent_++;

  if (schedule_ == ASYNCHRONOUS) {
    return PassAsynchronousMessage();
  } else if (schedule_ == RESIDUAL) {
    return PassResidualMessage();
  } else {
    CHECK(false, "invalid schedule");
    return true;
  }
}


double MessageQueue::Residual() {
  
  double residual = 0;
  for (int i = 0; i < static_cast<int>(message_heap_->size()); i++) {
    Message *message = message_heap_->at(i);
    message->ComputeMessage();
    residual += (message->NewValue() - message->CurrentValue()) *
      (message->NewValue() - message->CurrentValue());
  }
  return residual;
}


bool MessageQueue::PassAsynchronousMessage() {
  Message *message = message_heap_->at(asynch_pos_in_heap_);
  message->ComputeMessage();
  if (asynch_pos_in_heap_ % 1000000 == 0) {
    //cout << "Message: " << *message << endl;
  }

  if (asynch_pos_in_heap_ == static_cast<int>(message_heap_->size()) - 1) {
    double residual = Residual();
    cout << "Residual " << residual << endl;
    if (residual < .00000001) {
      //return true;
    }
  }

  message->UpdateMessage();

  //PushMessagesThroughVariables(message);

  // Increment the message we're on
  asynch_pos_in_heap_ = (asynch_pos_in_heap_ + 1) % message_heap_->size();

  return false;
}


void MessageQueue::PushMessagesThroughVariables(Message *message) {

  CHECK_EQ(message->Direction(), TO_VARIABLE, "Direction doesn't make sense");

  BinaryVariable *variable = message->GetEdge()->GetVariable();
  
  // Compute new message values for all edges leaving the variable other than
  // the edge that was just updated.
  for (uint i = 0; i < variable->Edges()->size(); i++) {
    Edge *v2f_edge = variable->Edges()->at(i);
    
    // Skip singletons
    if (v2f_edge->GetFactor()->Edges()->size() == 1) continue;  
    
    if (v2f_edge->PositionInVariable() != message->PositionInVariable()) {
      
      Message *v2f_message = v2f_edge->GetMessage(TO_FACTOR);
      
      if (v2f_message == NULL) {
	cout << v2f_edge->GetFactor()->Name() << endl;
	cout << v2f_edge->GetVariable()->Name() << endl;

	CHECK(false, "NULL message");
      }

      v2f_message->ComputeMessage();
      v2f_message->UpdateMessage();
	
      if (debug_) {
	cout << "   -> Updating " << *v2f_message << endl;
      }

      Factor *one_away_factor = v2f_edge->GetFactor();
      if (debug_) {
	cout << "    Computing all " << one_away_factor->Name() 
	     << " outgoing messages" << endl;
      }

      int pos_in_factor = v2f_message->PositionInFactor();
      one_away_factor->ComputeAllOutgoingMessagesExcept(pos_in_factor);

      for (uint j = 0; j < one_away_factor->Edges()->size(); j++) {
	if (static_cast<int>(j) == pos_in_factor) continue;
	
	Edge *f2v_edge = one_away_factor->Edges()->at(j);

	Message *f2v_message = f2v_edge->GetMessage(TO_VARIABLE);
	if (debug_) {
	  cout << "  [inner] updating priority " << *f2v_message << endl;
	}
	UpdateMessagePriority(f2v_message);


      }
    }
  } 
}


bool MessageQueue::PassResidualMessage() {

  Message *top_message = TopMessage();
  
  if (debug_ || messages_sent_ % (1 + size()  / 10000) == 1) {
    cout << messages_sent_ << " message: " << " " << *top_message << "   priority=" 
	 << top_message->Priority() <<  endl;  
  }
  if (messages_sent_ > 100 && top_message->Priority() < .000001) {
    cout << "CONVERGED" << endl;
    return true;
  }
  //for (uint i = 0; i < message_heap_->size(); i++) {
  //  cout << "   " << i << "=>" << *message_heap_->at(i) << endl;
  //}

  top_message->UpdateMessage();
  top_message->ComputeMessage();
  UpdateMessagePriority(top_message);


  if (messages_passed_.find(top_message->ID()) != messages_passed_.end()) {
    messages_passed_[top_message->ID()]++;
  } else {
    messages_passed_[top_message->ID()] = 1;
  }


  // We've made the variable dirty (by calling Message::UpdateMessage()), but we need 
  // to compute new outgoing messages from the variable so we know how to
  // update the message queue.
  if (true) {
    PushMessagesThroughVariables(top_message);

  } else if (top_message->Direction() == TO_VARIABLE) {
    BinaryVariable *variable = top_message->GetEdge()->GetVariable();

    // Compute new message values for all edges leaving the variable other than
    // the edge that was just updated.
    for (uint i = 0; i < variable->Edges()->size(); i++) {
      Edge *edge = variable->Edges()->at(i);

      // Skip singletons
      if (edge->GetFactor()->Edges()->size() == 1) {
	continue;	
      }

      if (edge->PositionInVariable() != 
	  top_message->PositionInVariable()) {

	Message *message = edge->GetMessage(TO_FACTOR);

	message->ComputeMessage();
	message->UpdateMessage();
	if (debug_) {
	  cout << "   -> updating " << *message << endl;
	}


	//UpdateMessagePriority(message);

	
	Factor *factor = edge->GetFactor();
	factor->ComputeAllOutgoingMessages();
	for (uint j = 0; j < factor->Edges()->size(); j++) {
	  if (factor->Edges()->at(j)->PositionInFactor() != 
	      message->PositionInFactor()) {

	    Message *message2 = factor->Edges()->at(j)->GetMessage(TO_VARIABLE);
	    //message2->ComputeMessage();
	    if (debug_) {
	      cout << "       -> updating " << *message2 << endl;
	    }
	    UpdateMessagePriority(message2);
	  }
	}
      }
      
      //UpdateMessagePriority(message);
    } 
    
  } else if (top_message->Direction() == TO_FACTOR) {
    CHECK(false, "Unexpected");
    Factor *factor = top_message->GetEdge()->GetFactor();

    for (uint i = 0; i < factor->Edges()->size(); i++) {
      Edge *edge = factor->Edges()->at(i);

      if (edge->PositionInFactor() != top_message->GetEdge()->PositionInVariable()) {
	Message *message = edge->GetMessage(TO_VARIABLE);

	message->ComputeMessage();
	UpdateMessagePriority(message);
      }
    }
  } else {
    SHOW(top_message->Direction());
    CHECK(false, "Unknown message direction");

  }

  if (schedule_ == RESIDUAL) {
    if (TopMessage()->Priority() <= 0.000001) {
      cout << "*** Messages converged ***" << endl;
      return true;  // converged
    } else {
      return false;  // not yet converged
    }
  } else {
    return false;
  }
}
