#include "sequence_factor.h"

#include <iostream>
#include <stdio.h>
#include "utils.h"

#include "binary_variable.h"

using namespace std;


SequenceFactor::SequenceFactor(string name) : Factor(name) {
  N_ = 0;

  lr_ = NULL;
  rl_ = NULL;
  forced_zeros_ = NULL;
  forced_ones_ = NULL;

}

SequenceFactor::~SequenceFactor() {
  delete incoming_messages_;

  delete lr_;
  delete rl_;

  delete forced_zeros_;
  delete forced_ones_;
}


void SequenceFactor::LoadIncomingMessagesFromEdges() {
  CHECK_EQ(edges_->size(), incoming_messages_->size(),
	   "Edges and incoming messages of different size");
  for (uint i = 0; i < edges_->size(); i++) {
    SetIncomingMessage(i, edges_->at(i)->ToFactor());
  }
}


void SequenceFactor::SetIncomingMessage(int var, double val) {
  // Check to see if we're changing the message value before we make the
  // factor dirty.  Later, we might want to check to see if the difference 
  // is within some tolerance.
  if (val != incoming_messages_->at(var)) {
    CHECK_SMALL(val, "Huge incoming message value");
    incoming_messages_->at(var) = val;
    Touch();
  }
}

double SequenceFactor::Likelihood() {
  int num_transitions = 0;

  bool last_var_on = false;
  for (uint i = 0; i < scope_->size(); i++) {
    BinaryVariable *var = scope_->at(i);
    CHECK(var->IsAssigned(), "Variable not assigned");
    if (var->Assignment()) {
      if (!last_var_on) {
	num_transitions++;
	last_var_on = true;
      }
    } else if (last_var_on) {
      num_transitions++;
      last_var_on = false;
    }
  }

  if (num_transitions <= 2) {
    return 0;
  } else {
    return -100000;
  }
}



double SequenceFactor::message(int var) {
  return forced_ones_->at(var) - forced_zeros_->at(var);
}

void SequenceFactor::ComputeOutgoingMessage(int position_in_factor) {
  //EXPECT(false, "Inefficient call to SequenceFactor::ComputeOutgoingMessage(int)");

  ComputeAllOutgoingMessages();
}



void SequenceFactor::ComputeAllOutgoingMessages() {

  LoadIncomingMessagesFromEdges();
  N_ = static_cast<int>(edges_->size());
  
  if (lr_ == NULL) {
    lr_ = new vector<double>(N_);
    rl_ = new vector<double>(N_);
  }
  if (forced_ones_ == NULL) {
    forced_zeros_ = new vector<double>(N_);
    forced_ones_ = new vector<double>(N_);
  }

  for (int i = 0; i < N_; i++) {
    lr_->at(i) = 0;
    rl_->at(i) = 0;
    forced_zeros_->at(i) = 0;
    forced_ones_->at(i) = 0;
  }

  CHECK_EQ(lr_->size(), incoming_messages_->size(),
	   "LR vector not the right size in SequenceFactor");
  CHECK_EQ(rl_->size(), incoming_messages_->size(),
	   "RL vector not the right size in SequenceFactor");
  CHECK_EQ(forced_zeros_->size(), incoming_messages_->size(),
	   "LR vector not the right size in SequenceFactor");
  CHECK_EQ(forced_ones_->size(), incoming_messages_->size(),
	   "RL vector not the right size in SequenceFactor");

  // Make left-to-right and right-to-left passes
  for (int i = 0; i < N_; i++) {
    if (i == 0) {
      lr_->at(i) = incoming_messages_->at(i);
      rl_->at(N_-1) = incoming_messages_->at(N_-1);
    } else {
      lr_->at(i) = MAX(lr_->at(i-1), 0) + incoming_messages_->at(i);
      rl_->at(N_-i-1) = MAX(rl_->at(N_-i), 0) + incoming_messages_->at(N_-i-1);
    }
  }

  // Take the sum of the left side and right side, constrained
  // to use the adjacent h.
  for (int i = 0; i < N_; i++) {
    if (i > 0) {
      forced_ones_->at(i) += MAX(lr_->at(i-1), 0);
    }
    
    if (i < N_ - 1) {
      forced_ones_->at(i) += MAX(rl_->at(i+1), 0);
    }
  }
 
  // Take cumulative maxes
  for (int i = 0; i < N_; i++) {
    if (i > 0) {
      lr_->at(i) = MAX(lr_->at(i-1), lr_->at(i));
      rl_->at(N_-i-1) = MAX(rl_->at(N_-i), rl_->at(N_-i-1));
    }
  }

  // Take either the left side max or the right side max
  // (assuming we're not at the edge)
  for (int i = 0; i < N_; i++) {
    double val = 0;
    if (i > 0) {
      val = MAX(val, lr_->at(i-1));
    }
    if (i < N_ - 1) {
      val = MAX(val, rl_->at(i+1));
    }
    forced_zeros_->at(i) = val;
  }

  for (int i = 0; i < N_; i++) {
    outgoing_messages_->at(i) = forced_ones_->at(i) - forced_zeros_->at(i);
  }

  bool PRINT_DEBUG = false;

  if (PRINT_DEBUG) {
    cout << "Internal SequenceFactor caches (size=" << N_ << ")" << endl;
    for (int i = 0; i < N_; i++) {
      cout << lr_->at(i) << " ";
    }
    cout << endl;
    for (int i = 0; i < N_; i++) {
      cout << rl_->at(i) << " ";
    }
    cout << endl;

    for (int i = 0; i < N_; i++) {
      cout << forced_zeros_->at(i) << " ";
    }
    cout << endl;
    for (int i = 0; i < N_; i++) {
      cout << forced_ones_->at(i) << " ";
    }
    cout << endl;
  }
}


/*
int test_sequence_factor(char *argv, int argc) {

  int N = 1000;
  SequenceFactor *seq = new SequenceFactor(N);


  double weights[N];
  double forced_zeros[N];
  double forced_ones[N];

  for (int i = 0; i < N; i++) {
    double x = i / 2.0 - 20.0;
    weights[i] = x * x / 100 - 5 * (i - 50) + 10;
  }

  cout << "Weights:" << endl;
  for (int i = 0; i < N; i++) {
    cout << weights[i] << " ";
  }
  cout << endl << endl;

  seq->set_incoming_messages(weights);

  seq->compute_all_messages();

  for (int i = 0; i < N; i++) {
    cout << seq->message(i) << " ";
  }
  cout << endl;

}
*/
